# Color Channel Image Reconstructor 32/16/8bit Version
# Ver. 0.1
# add 32bit support, and modification for ImageJ Ver. 1.54 (0.1) Aug. 2023
# Silent & Verbose Version
# (c)2021-3 Ohnishi, Yasuo
from ij import IJ 
from ij import plugin
from ij.io import OpenDialog
from ij.gui import GenericDialog
from ij.plugin import ChannelSplitter
from ij.plugin import Duplicator
from ij.process import ImageConverter
from loci.plugins import BF
import os
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

def convValue(Value, L_max):
#   Converting Luminance Value each in 8, 16, 32bit mode
    if L_max == 65535:
        Value = Value * 256
    elif L_max == 16777215:
        Value = Value * 256 * 256
    elif L_max == 1.0:
        Value = Value / 255.0
    else:
        Value = Value
    return Value
#--------------------------
def set_Lmax(imp):
    # set Maximum value of Lightness
    bitdepth = imp.getBitDepth()
    if bitdepth == 32:
        L_max = imp.getDisplayRangeMax()
        if int(L_max) > 65536:
            L_max = 16777215
        elif int(L_max) > 256:
            L_max = 65535
        elif int(L_max) > 1.01:
            L_max = 255
        else:
            L_max = 1.0
    elif bitdepth == 16:
        L_max = 65535
    else: # 8 or 24 bit
        L_max = 255
    return L_max
#-------------------------
def allocateCh(chName, impR, impG, impB, impGray):
    if chName == "R":
        imporg = impR
        imp = Duplicator().run(impR)
    elif chName == "G":
        imporg = impG
        imp = Duplicator().run(impG)
    elif chName == "B":
        imporg = impB
        imp = Duplicator().run(impB)
    else:
        imporg = impGray
        imp = Duplicator().run(impGray)
    return imp, imporg

#-------------------------
def mixCh(MixRatio, MixMode, Gamma, imp, impMx, impMxOrg, L_max):
    #  channel mixing routine
    # MixRatio: mixing ratio of mixing ch
    # MixMode: mixing mode of mixing ch darken / normal /lighten
    # Gamma: gamma correction value of mic ch.
    # imp: original ch
    # impMx: mixing ch
    # impMxOrg: origical mixing ch (not modified)
    # L_max: max value
    strRatio = str(MixRatio)
    IJ.run(impMx, "Multiply...", "value=" + strRatio)
    #  Multiply Gammma for 2nd Mixing Ch.
    impMx.setDisplayRange(0, L_max)
    IJ.run(impMx, "Gamma...", "value=" + str(Gamma))
    impMx.setDisplayRange(0, L_max)
    #  Mix imp11 and 2nd Mixing Channel
    ic = plugin.ImageCalculator()
    impMxd0 = ic.run("Add create", imp, impMx) # No window made
    if MixMode == "Darken":
        ic = plugin.ImageCalculator()
        impMxd1 = ic.run("min create", impMxd0, impMxOrg) # No window made
        impMxd0.close
    elif MixMode == "Lighten":
        ic = plugin.ImageCalculator()
        impMxd1 = ic.run("max create", impMxd0, impMxOrg) # No window made
        impMxd0.close
    else:
        impMxd1 = impMxd0
    return impMxd1

#####################################################
def RGBmanipulation(folder,filename,parm):
#  input Parameters
    if parm is not None:
# if parm is not null object, then...
        Mode, BaseCh, MixCh, MixRatio, MixMode, Show, GIMPout, Gamma, \
                                MixCh1, MixRatio1, MixMode1, Gamma1 = parm
        print(Mode, BaseCh, MixCh, MixRatio, MixMode, Show, GIMPout, Gamma) 
#     Mode: Image bit depth 32-bit, 16-bit or 8-bit
#     Show: Show working windows or not
#     BaseCh: original ch.
#     MixCh: 1st mixing ch.
#     MixCh1: 2nd mixing ch.
#     MixRatio: mixing ratio of 1st mixing ch.
#     MixRatio1: mixing ratio of 2nd mixing ch.
#     Gamma, Gamma1: gamma value  of 1st & 2nd mixing ch.
#     GIMPout: use file name for GIMP editing? True / False

    # Always Open file with Bio-format Plugin
    imps = BF.openImagePlus(folder + filename) # Open file with Bio-format Plugin
    imp0 = imps[0] #imp0: Original Image 
    if Mode == "Auto":
        bitdepth = imp0.getBitDepth()
        if bitdepth == 8 or bitdepth == 24:
            Mode = "8-bit"
        elif bitdepth ==16:
            Mode = "16-bit"
        elif bitdepth ==32:
            Mode = "32-bit"
        else:
            Mode = "16-bit"
    L_max = set_Lmax(imp0)
    if Show == "Yes":
    # Windows Showing Mode
        imp0.show()
    #------
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0
    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        else:
            ImageConverter(i).convertToGray32()
        i.setDisplayRange(0, L_max)

    impGray = Duplicator().run(imp0)
    if MixCh == "Luminosity" or  MixCh1 == "Luminosity":
        IJ.run(impGray, "Conversions...", "weighted")
    if Mode == "32-bit":
        ImageConverter(impGray).convertToGray32()
    elif Mode == "16-bit":
        ImageConverter(impGray).convertToGray16()
    elif Mode == "8-bit":
        ImageConverter(impGray).convertToGray8()
    impGray.setDisplayRange(0, L_max)

    ## R ch.
    impR =  impsCh[0]
    IJ.saveAsTiff(impR, folder + filename +"_R.tif")
    ## G ch.
    impG =  impsCh[1]
    IJ.saveAsTiff(impG, folder + filename +"_G.tif")
    ## B ch. & B x 0.3
    impB =  impsCh[2]
    IJ.saveAsTiff(impB, folder + filename +"_B.tif")
    # Show Windows
    imp0.close()
    if Show == "Yes":
        impR.show()
        impG.show()
        impB.show()
    imp1, imp1org = allocateCh(BaseCh, impR, impG, impB, impGray)
    # 1st Mixing Ch.
    imp2, imp2org = allocateCh(MixCh, impR, impG, impB, impGray)
    # 2nd Mixing Ch.
    imp3, imp3org = allocateCh(MixCh1, impR, impG, impB, impGray)
    if MixCh1 == "None":
        MixRatio1 = 0.0
    ####  Convert Base Channel
    Ratio1 = str(1.0 - (MixRatio + MixRatio1))
    IJ.run(imp1, "Multiply...", "value=" + Ratio1)
    if Show == "Yes":
        imp1.show()
    ######### Mixing 1st Channel #############
    if Show == "Yes":
        imp2.show()
    imp11 = mixCh(MixRatio, MixMode, Gamma, imp1, imp2, imp2org, L_max)
    # output for 1st Mixed Ch
    if MixCh1 == "None":
        IJ.saveAsTiff(imp11, folder + filename +"_" + BaseCh + "x" + \
            Ratio1 + "_" + MixCh + "x" + str(MixRatio) + MixMode + \
            "Gm" + str(Gamma) + ".tif")
        if GIMPout =="Yes":
            IJ.saveAsTiff(imp11, folder + filename +"_Ch_Reconstructed.tif")
    else: 
    ######### Mixing 2nd Channel #############
        imp21 = \
            mixCh(MixRatio1, MixMode1, Gamma1, imp11, imp3, imp3org, L_max)
    # output for 2nd Mixed Ch
        IJ.saveAsTiff(imp21, folder + filename +"_" + BaseCh + "x" + \
            Ratio1 + "_" + MixCh + "x" + str(MixRatio) + MixMode + \
            "Gm" + str(Gamma) +  "_" + MixCh1 + "x" + str(MixRatio1) + \
            MixMode1 + "Gm" + str(Gamma1) + ".tif")
        if GIMPout =="Yes":
            IJ.saveAsTiff(imp21, folder + filename +"_Ch_Reconstructed.tif")
        if MixMode1 != "Normal":
            imp21.hide()
            imp21.close()

    if MixMode != "Normal":
        imp11.hide()
        imp11.close()
    if MixCh1 != "None":
        imp3.hide()
        imp3.close()
    imp1.hide()
    imp2.hide()
    impR.hide()
    impG.hide()
    impB.hide()
    impGray.hide()
    imp1.close()
    imp2.close()
    impR.close()
    impG.close()
    impB.close()
    impGray.close()
#####################################################
# Get parameters from the input DialogBox 
def getParameters():
    gd = GenericDialog("Color Channel Reconstructor / Enter Parameters")
    selectBaseCh = ["R","G","B"]
    gd.addChoice("Channel for Reconstructuon", selectBaseCh, selectBaseCh[1])
#
    gd.addMessage("[1st Mixing Channel]")
    selectMixCh = ["R","G","B","Luminosity", "Lightness"]
    gd.addChoice("Mixing Channel", selectMixCh, selectMixCh[0])
    gd.addNumericField("Channel Mixing Ratio (0 to 1.0)", 0.5, 1)
    selectAdjst = ["Darken","Normal","Lighten"]
    gd.addChoice("Mixing Mode", selectAdjst, selectAdjst[1])
    gd.addNumericField("Gamma for Mixing Ch.(0 to 5.0)", 1.0, 1)
#
    gd.addMessage("[2nd Mixing Channel]")
    selectMixCh1 = ["None","R","G","B","Luminosity", "Lightness"]
    gd.addChoice("Mixing Channel", selectMixCh1, selectMixCh1[0])
    gd.addNumericField("Channel Mixing Ratio (0 to 1.0)", 0.0, 1)
    selectAdjst1 = ["Darken","Normal","Lighten"]
    gd.addChoice("Mixing Mode", selectAdjst1, selectAdjst1[1])
    gd.addNumericField("Gamma for Mixing Ch.(0 to 5.0)", 1.0, 1)
#
    gd.addMessage("    ")
    selectGIMPoutput = ["Yes","No"]
    gd.addChoice("Output File name for GIMP plugin?", selectGIMPoutput, selectGIMPoutput[0])
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])
    gd.showDialog()
    Mode   = str("Auto")
    BaseCh   = str(gd.getNextChoice())
# get 1st Mixing Ch.
    MixCh   = str(gd.getNextChoice())
    MixRatio  = gd.getNextNumber()
    MixMode   = str(gd.getNextChoice())
    Gamma =  gd.getNextNumber()
# get 2nd Mixing Ch.
    MixCh1   = str(gd.getNextChoice())
    MixRatio1  = gd.getNextNumber()
    MixMode1   = str(gd.getNextChoice())
    Gamma1 =  gd.getNextNumber()

    GIMPout   = str(gd.getNextChoice())
    Show   = str(gd.getNextChoice())
    if MixRatio > 1.0:
        MixRatio = 1.0
    elif MixRatio < 0.0:
        MixRatio = 0.0
    if Gamma > 5.0:
        Gamma = 5.0
    elif Gamma < 0.0:
        Gamma = 0.0

    if MixRatio + MixRatio1 > 1.0:
        MixRatio1 = 1.0 - MixRatio
        IJ.showMessage("2nd Mixing Ratio was changed to " + str(MixRatio1))
    elif MixRatio1 < 0.0:
        MixRatio1 = 0.0
    if Gamma1 > 5.0:
        Gamma1 = 5.0
    elif Gamma1 < 0.0:
        Gamma1 = 0.0
    if MixCh == "Luminosity" and MixCh1 == "Lightness":
        IJ.showMessage("2nd Lightness Ch. will be changed to None!")
        MixCh1 = "None"
    elif MixCh1 == "Luminosity" and MixCh == "Lightness":
        IJ.showMessage("2nd Luminosity Ch. will be changed to None!")
        MixCh1 = "None"
    if gd.wasCanceled():  
        Mode   = "Stop"
        IJ.log("Canceled")
    if BaseCh == MixCh:
        IJ.showMessage("Mixing Channel is same with Reconstructing Channel!")
#        Mode   = "Stop"
    return Mode, BaseCh, MixCh, MixRatio, MixMode, Show, GIMPout, Gamma, MixCh1, MixRatio1, MixMode1, Gamma1
##############
# Main
od = OpenDialog("Choose a file")
folder = od.getDirectory()
file = od.getFileName()
path = str(folder) + str(file)
IJ.log(path)
IJ.log(folder)
IJ.log(file)
parm = getParameters()
print ("Parm[0]=", parm[0])
if parm[0] != "Stop":
    RGBmanipulation(folder,file,parm)
    IJ.log("Finish")
#    IJ.run("Quit")
